freeze;

declare attributes FldFun: PythNum, PythNumElt, Hght;
declare attributes DiagQuadFrm: primes;

// TODO: LegendreSymbol and IsLocalSquare in characteristic 2

ZZ := Integers();
QQ := Rationals();

//////////////////////////////////////////////////
// LegendreSymbol

intrinsic LegendreSymbol( a::FldFunElt, P::PlcFunElt ) -> RngIntElt
{ 
Given an element a and a place P of a function field, 
this function computes the Legendre symbol of a square-class of a over P. 
}
    K := FunctionField(P);
    require Characteristic(K) ne 2: "Implemented only for fields of odd characteristic!";
    require IsCoercible(K, a) : "Incompatible fields!";
    require IsFinite(ConstantField(K)): "The base field must be finite!";

    v := Valuation(a, P);
    if IsOdd(v) then
        return 0;
    end if;
    if v ne 0 then
        p := LocalUniformizer(P);
        a := a*p^(-v);
    end if;
    if IsSquare(Evaluate(a,P)) then
        return 1;
    end if;
    return -1;
end intrinsic;

intrinsic LegendreSymbol( a::FldFunElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return LegendreSymbol(a, Place(P));
end intrinsic;


//////////////////////////////////////////////////
// IsLocalSquare
intrinsic IsLocalSquare( a::FldFunElt, P::PlcFunElt ) -> BoolElt
{
Given an element a in K and a place P of a function field, 
checks whether a is a square at the completion of K at P.
}
    return LegendreSymbol(a,P) eq 1;
end intrinsic;

//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldFunElt ) -> BoolElt
{ " }
    K := Parent(a);
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    return true;
end intrinsic;

intrinsic IsSOS( a::FldFunElt ) -> BoolElt
{ " }
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a global function field K and a place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

intrinsic LengthOfSOS( a::FldFunElt ) -> RngIntElt
{
Given a nonzero element a of a global function field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    K := Parent(a);
    // find all the primes of K that matter for a
    primes := Support(Divisor(a));
    vprint DQF_verbosity : "Primes to be tested:", primes;
    if IsEmpty(primes) then
        // a is a constant
        return LengthOfSOS( ConstantField(K)!a );
    end if;

    n := Max([ length_nondyadic(a, P) : P in primes ]);
    return n;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldFunElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// Level
// (just for completeness)
intrinsic Level( K::FldFun ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    return LengthOfSOS( K!-1);
end intrinsic;


//////////////////////////////////////////////////
// IsFormallyReal
// just for completeness
intrinsic IsFormallyReal( K::FldFun ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    return IsSOS(K!-1);
end intrinsic;

intrinsic IsReal( K::FldFun ) -> BoolElt
{ " } //"
    return IsFormallyReal(K);
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber & PythagorasElement

intrinsic PythagorasNumber( K::FldFun ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";
    if assigned K`PythNum then
        return K`PythNum;
    end if;
    
    // there are no dyadic primes, hence
    // P(K) must be either 2 or 3, depending on s(K)
    K`PythNum := IsSquare(K!-1) select 2 else 3;
    return K`PythNum;
end intrinsic;

intrinsic PythagorasElement( K::FldFun ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";
    if assigned K`PythNumElt then
        return K`PythNumElt;
    end if;
    
    PK := PythagorasNumber(K);
    sK := Level(K);
    if sK eq 1 then
        // PK == 2
        // find any non-square
        // we are in a global case, hence there is an odd degree place
        // so any nonsquare in the field of constants is fine
        F, m := ExactConstantField(K);
        K`PythNumElt := m(Nonsquare(F));
        assert2 not IsSquare(K`PythNumElt);
    else
        // PK == 3
        // find a place where -1 is not a square
        // it must have an odd degree 
        S := Support( DivisorOfDegreeOne(K) );
        for p in S do
            if IsOdd(Degree(p)) then
                K`PythNumElt := LocalUniformizer(p);
                break;
            end if;
        end for;
    end if;
    return K`PythNumElt;
end intrinsic;


//////////////////////////////////////////////////
// HilbertSymbol 
// (partially based on Magma code for number fields)

intrinsic HilbertSymbol( a::FldFunElt, b::FldFunElt, P::PlcFunElt ) -> RngIntElt
{
Given a,b in a global function field and a prime p of K, 
returns the Hilbert symbol (a,b)_P.
}
    require Parent(a) eq Parent(b) : "Arguments must belong to the same field!";
    K := FunctionField(P); 
    require IsCoercible(K, a) and IsCoercible(K, b) : "Incompatible fields!";
    require not IsZero(a) and not IsZero(b): "Arguments must be non zero!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";

    va := Valuation(a, P);
    vb := Valuation(b, P);
    p := LocalUniformizer(P);
    a_ := a*p^(-va);
    b_ := b*p^(-vb);
    
    // four cases two consider
    if IsEven(va) and IsEven(vb) then 
        return 1;
    elif IsEven(va) and IsOdd(vb) then
        return IsSquare(Evaluate(a_,P)) select 1 else -1;
    elif IsOdd(va) and IsEven(vb) then
        return IsSquare(Evaluate(b_,P)) select 1 else -1;
    else
        return IsSquare(Evaluate(-a_*b_,P)) select 1 else -1;
    end if;
end intrinsic;

intrinsic HilbertSymbol( a::FldFunElt, b::FldFunElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HilbertSymbol(a, b, Place(P));
end intrinsic

//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{
The Hasse invariant of a diagonal quadratic form q = < a1, ..., ad> \
at a prime P is a product of Hilbert symbols (c.f. [Lam (2005), Def. V.3.17]) \
prod_(i<j) (ai, aj)_p.
This function computes the Hasse symbol of q at p.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HasseMinkowskiInvariant(q, Place(P));
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HasseMinkowskiInvariant(q, Place(P));
end intrinsic;
    
//////////////////////////////////////////////////
// IsIsotropic

intrinsic IsIsotropic( q::DiagQuadFrm, P::PlcFunElt ) -> BoolElt
{
Checks if a quadratic form q over a global function field is 
locally isotropic at P.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";

    // unary forms cannot be isotropic
    if Dimension(q) le 1 then
        return false;
    end if;
    // quintic and higher dimensional forms are always isotropic
    if Dimension(q) ge 5 then
        return true;
    end if;

    // the base field and its maximal order
    K := q`base_ring;
  
    // the residue field
    Kp, kappa := ResidueClassField(P);
    // uniformizer
    p := LocalUniformizer(P);

    // split the form into q = q0 + p*q1
    q0 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsEven(Valuation(a,P)) ];
    q1 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsOdd(Valuation(a,P)) ];
    //q0 := [ a*Denominator(a)^2 : a in q0 ];
    //q1 := [ a*Denominator(a)^2 : a in q1 ];
    
    // take the reduction mod gp
    q0bar := [ Evaluate(a, P) : a in q0 ];
    q1bar := [ Evaluate(a, P) : a in q1 ];
    // verify if q0, q1 are isotropic
    A := IsIsotropic( DiagonalQuadraticForm(Kp, q0bar) );
    B := IsIsotropic( DiagonalQuadraticForm(Kp, q1bar) );
    return A or B;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngFunOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return IsIsotropic(q, Place(P));
end intrinsic;

function is_isotropic_GFF( q )
    // Given a quadratic form q with coefficients in a global function field K,
    // this function tests if q is isotropic.

    // the base field
    K := q`base_ring;

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        t := IsSquare(Discriminant(q));
        return t;
    end if;

    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
        q`primes := [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // check that q is locally isotropic over all of them
    return forall{ P : P in q`primes | IsIsotropic(q, P) };
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::PlcFunElt ) -> BoolElt
{
Given a quadratic form q over a global function field K and a prime P of K, 
this function returns true iff q is hyperbolic in the completion
of K at P.
}
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(K!-1,K!-1,P)^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngFunOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    return IsHyperbolic(q, Place(P));
end intrinsic;

function is_hyperbolic_GFF( q )
    // Given a quadratic form q over a number field K, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    K := q`base_ring;

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
        q`primes := [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // check that q is locally isotropic over all of them
    return forall{ P : P in q`primes | IsHyperbolic(q, P) };
end function;


//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-dimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } // "
    require IsPrime(P): "P must be a prime ideal!";
    return AnisotropicDimension(q, Place(P));
end intrinsic;

function anisotropic_dimension_GFF( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a global function field K, 
// this function returns a dimension of an anisotropic part of q.

    // the base field
    K := q`base_ring;

    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
        q`primes := [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // compute maximal anisotropic dimension over all primes
    M := Max([ AnisotropicDimension(q, P) : P in q`primes ]);
    return M;
end function;


//////////////////////////////////////////////////
// IsTorsion
// just for completeness
function is_torsion_GFF( _q )
// Check if the element _q of a Witt ring is torsion.
    return true;
end function;


//////////////////////////////////////////////////
// AreWittEquivalent

intrinsic AreWittEquivalent( K::FldFun, L::FldFun ) -> BoolElt
{
Given two fields K and L, this function return true if their Witt rings are isomorphic and false otherwise.
}    
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.1
    if Characteristic(K) eq 2 and Characteristic(L) ne 2 then
        return false;
    elif Characteristic(K) ne 2 and Characteristic(L) eq 2 then
        return false;
    elif Characteristic(K) eq 2 and Characteristic(L) eq 2 then
        return true;
    else
        // Now both must have odd characteristic
        // Use Szymiczek (1991), Theorem 1.3
        return Level(K) eq Level(L);
    end if;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.1
    if Characteristic(K) eq 2 and Characteristic(L) ne 2 then
        return false;
    elif Characteristic(K) ne 2 and Characteristic(L) eq 2 then
        return false;
    elif Characteristic(K) eq 2 and Characteristic(L) eq 2 then
        return true;
    else
        // Now both must have odd characteristic
        // Use Szymiczek (1991), Theorem 1.3
        return Level(K) eq Level(L);
    end if;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldNum ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;




