freeze;

declare attributes DiagQuadFrm: primes;

import "cqf_FldNum.m" : is_locally_isotropic_nondyadic;

ZZ := Integers();
QQ := Rationals();

//////////////////////////////////////////////////
// IsLocalSquare

intrinsic IsLocalSquare( a::FldRatElt, P::RngInt ) -> BoolElt
{ 
Given an element a in QQ, decide if a is a square in the completion QQ_P
}
    return IsLocalSquare(a, Generator(P));
end intrinsic;

//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldRatElt ) -> BoolElt
{
Returns true iff a is a sum of squares
}
    return a ge 0;
end intrinsic;

intrinsic IsSumOfSquares( a::RngIntElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;

intrinsic IsSOS( a::FldRatElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;

intrinsic IsSOS( a::RngIntElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

intrinsic LengthOfSOS( a::RngIntElt ) -> RngIntElt
{
Given a nonzero integer a, this fuction determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares.
}
    require not IsZero(a): "The argument must be nonzero!";
    
    // start with trivial cases
    if a lt 0 then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // special case
    if a eq 2 then
        return 2;
    end if;
    
    // check if a is a sum of 3 squares
    // write a in a form a = 4^v'*a'
    v := Valuation(a, 2);
    v_ := Floor( v/2 );
    a_ := ZZ ! ( a*4^(-v_) );
    // use Lagrange 3-square theorem
    if (a_ mod 8) eq 7 then
        return 4;
    end if;
    
    // now a_ (hence a) is a sum of 2 or 3 squares
    // compute the radical of a
    m, _ := SquarefreeFactorization(a);
    // use Lagrange 2-square theorem
    sq2 := forall{ p : p in PrimeDivisors(m) | (p mod 4) eq 1 };
    return sq2 select 2 else 3;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::RngIntElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;

intrinsic LengthOfSOS( a::FldRatElt ) -> RngIntElt
{
Given a nonzero rational number a, this fuction determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares.
}
    return LengthOfSOS( Numerator(a)*Denominator(a) );
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldRatElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// IsFormallyReal
// just for completeness
intrinsic IsFormallyReal( K::FldRat ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    return true;
end intrinsic;

intrinsic IsReal( K::FldRat ) -> BoolElt
{ " } //"
    return true;
end intrinsic;


//////////////////////////////////////////////////
// Level
// just for completeness
intrinsic Level( K::FldRat ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    return 0;
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber
// just for completeness
intrinsic PythagorasNumber( K::FldRat ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    return 4;
end intrinsic;

intrinsic PythagorasElement( K::FldRat ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    return 7;
end intrinsic;

//////////////////////////////////////////////////
// Height
// just for completeness
intrinsic Height( K::FldRat ) -> RngIntElt
{
This function computes the height h(K) of a number field K,
i.e. exp( Wt(K) ), where Wt(K) is denotes the torsion subgroup
of the Witt ring of K..
}
    return 4;
end intrinsic;




//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;


//////////////////////////////////////////////////
// IsIsotropic

function is_locally_isotropic_QQ2( q )
    // Given a quadratic form q over QQ 
    // this function checks if the localization of q at 2 is isotropic.

    // the dimension
    d := Dimension(q);
    // the determinant
    detq := &* q`coeffs;
    // only square class matters
    detq := Numerator(detq)*Denominator(detq);
    //QQ2 := pAdicField(2);
    
    if d le 1 then
        return false;
    elif d ge 5 then
        return true;
    elif d eq 2 then
        return IsLocalSquare(-detq, 2);
    elif d eq 3 then
        // use [Lam, Prop. V.3.22]
        s := HilbertSymbol(-1, -detq, 2);
        sq := WittInvariant(q, 2);
        return (s eq sq);
    elif IsLocalSquare(detq, 2) then
        // if we get here, then d == 4        
        // for quartic form with determinant 1, use [ Lam, Proposition V.3.23]
        s := HilbertSymbol(-1, -1, 2);
        sq := WittInvariant(q, 2);
        vprint DQF_verbosity: "s =", s, "sq =", sq;
        return (s eq sq);
    end if;
    // form is unique with det a square
    return true;
end function;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngInt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally isotropic at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    if (QQ!2) notin P then
        return is_locally_isotropic_nondyadic(q,P);
    else
        return is_locally_isotropic_QQ2(q,P);
    end if;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngIntElt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally isotropic at P.
}
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    if P ne 2 then
        return is_locally_isotropic_nondyadic(q,P*ZZ);
    else
        return is_locally_isotropic_QQ2(q);
    end if;
end intrinsic;

function is_isotropic_QQ( q )
    // Given a quadratic form q with coefficients in QQ,
    // this function tests if q is isotropic.

    // the base field and its maximal order
    assert q`base_ring eq QQ;

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        return IsSquare(Discriminant(q));
    end if;
    vprint DQF_verbosity : "General dimension.";

    // compute the signature of q
    if forall{ c : c in q`coeffs | c gt 0} or forall{ c : c in q`coeffs | c lt 0} then
        return false;
    end if;
    vprint DQF_verbosity : "Signature test passed.";
          
    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
        q`primes := [2] cat [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // check that q is locally isotropic over all of them
    vprint DQF_verbosity : "Local results:", [ <P, IsIsotropic(q, P)> : P in q`primes ];
    return forall{ P : P in q`primes | IsIsotropic(q, P) }; 
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngIntElt ) -> BoolElt
{
Checks if a quadratic form q over QQ is locally hyperbolic at P.
}
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(-1,-1,P)^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngInt ) -> BoolElt
{ " } //"
    return IsHyperbolic(q, Generator(P));
end intrinsic;

function is_hyperbolic_QQ( q )
    // Given a quadratic form q over rationals, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    //K := q`base_ring;
    //OK := MaximalOrder(K);

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // compute the signature of q
    if not IsZero( &+ [ Sign(c) : c in q`coeffs ] ) then
        return false;
    end if;
    vprint DQF_verbosity : "Signature test passed.";
          
    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
        q`primes := [2] cat [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // check that q is locally hyperbolic over all of them
    vprint DQF_verbosity : "Local results:", [ <P, IsHyperbolic(q, P)> : P in q`primes ];
    return forall{ P : P in q`primes | IsHyperbolic(q, P) };
end function;

//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngInt ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-fimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngIntElt ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime number!";
    require q`base_ring eq QQ: "Base field of P does not match base field of q!";
    return AnisotropicDimension(q, P*ZZ);
end intrinsic;

function anisotropic_dimension_QQ( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a number field K, 
// this function returns a dimension of an anisotropic part of q.

    // compute the signatures of q in all real embedding of K
    N := Abs( &+ [ Sign(c) : c in q`coeffs ] );
    vprint DQF_verbosity : "Signature: ", N;
    // for n > 2, it must be a real place, where the the anisotropic
    // part has a maximal dimension
    if N  ge 3 then
        return N;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
    // find all the primes of K that matter for q
    if not assigned q`primes then
        q`primes := &join[ Seqset(PrimeDivisors(Numerator(c)*Denominator(c))) : c in q`coeffs ];
        q`primes := [2] cat [ P : P in q`primes ];
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // compute maximal anisotropic dimension over all primes
    M :=  Max([ AnisotropicDimension(q, P) : P in q`primes ]);
    return Max(M, N); 
end function;


//////////////////////////////////////////////////
// IsTorsion

function is_torsion_QQ( _q )
// Check if the element _q of a Witt ring is torsion.
    q := _q`form;
    sgn_q := Abs( &+ [ Sign(c) : c in q`coeffs ] );
    return IsZero(sgn_q);
end function;

//////////////////////////////////////////////////
// AreWittEquivalent
// just for completeness
intrinsic AreWittEquivalent( K::FldRat, L::FldRat ) -> BoolElt
{
Given two number fields K and L, this function
return true if their Witt rings are isomorphic and false otherwise.
}    
    return true;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldNum ) -> BoolElt
{
Given two number fields K and L, this function
return true if their Witt rings are isomorphic and false otherwise.
}    
    return AbsoluteDegree(L) eq 1;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFun[FldFin] ) -> BoolElt
{ " } //"
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFun ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

